# Copyright (C) 2005  Network Applied Communication Laboratory Co., Ltd.
#
# This file is part of Rast.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 02111-1307, USA

require "test/unit"
require "fileutils"
require "bdb"

require "rast"
require "test-utility"
require "rast/database-generatable"

module Rast
  class DBTest < Test::Unit::TestCase
    include DatabaseGeneratable

    def test_create
      options = {
        "byte_order" => Rast::LITTLE_ENDIAN,
        "encoding" => "utf8",
        "preserve_text" => true,
        "properties" => [
          {
            "name" => "title",
            "type" => Rast::PROPERTY_TYPE_STRING,
            "search" => true,
            "text_search" => true,
            "full_text_search" => true,
            "unique" => false,
          },
        ],
      }
      db_name = generate_db_name
      DB.create(db_name, options)
      DB.open(db_name, Rast::DB::RDONLY) do |db|
        properties = db.properties
        assert_equal(Rast::LITTLE_ENDIAN, db.byte_order)
        assert_equal("utf8", db.encoding)
        assert_equal("title", properties[0].name)
        assert_equal(Rast::PROPERTY_TYPE_STRING, properties[0].type)
        assert_equal(true, properties[0].search)
        assert_equal(true, properties[0].text_search)
        assert_equal(true, properties[0].full_text_search)
        assert_equal(false, properties[0].unique)
        assert_equal(1, properties.length)
      end
      assert_equal(true, btree_db_exist?(db_name, "text.ngm"))
      assert_equal(true, btree_db_exist?(db_name, "text.rng"))
      assert_equal(true, file_exist?(db_name, "text.pos"))
      assert_equal(true, file_exist?(db_name, "text.pfl"))
      assert_equal(true, btree_db_exist?(db_name, "properties.db"))
      assert_equal(true, property_btree_db_exist?(db_name, "title.inv"))
      assert_equal(true, property_btree_db_exist?(db_name, "title.ngm"))
      assert_equal(true, property_btree_db_exist?(db_name, "title.rng"))
      assert_equal(true, property_file_exist?(db_name, "title.pos"))
      assert_equal(true, property_file_exist?(db_name, "title.pfl"))
      assert_equal(true, dir_exist?(db_name, "text"))
      check_doc_info(db_name, 0, 0)

      options = {
        "byte_order" => Rast::BIG_ENDIAN,
        "encoding" => "utf8",
        "preserve_text" => false,
        "properties" => [
          {
            "name" => "filename",
            "type" => Rast::PROPERTY_TYPE_STRING,
            "search" => true,
            "text_search" => true,
            "full_text_search" => true,
            "unique" => true,
          },
          {
            "name" => "filesize",
            "type" => Rast::PROPERTY_TYPE_UINT,
            "search" => true,
            "text_search" => false,
            "full_text_search" => false,
            "unique" => false,
          },
          {
            "name" => "date",
            "type" => Rast::PROPERTY_TYPE_DATE,
            "search" => true,
            "text_search" => false,
            "full_text_search" => false,
            "unique" => false,
          },
        ],
      }
      db_name = generate_db_name
      DB.create(db_name, options)
      DB.open(db_name, Rast::DB::RDONLY) do |db|
        properties = db.properties
        assert_equal(Rast::BIG_ENDIAN, db.byte_order)
        assert_equal("utf8", db.encoding)
        assert_equal("filename", properties[0].name)
        assert_equal(Rast::PROPERTY_TYPE_STRING, properties[0].type)
        assert_equal(true, properties[0].search)
        assert_equal(true, properties[0].text_search)
        assert_equal(true, properties[0].full_text_search)
        assert_equal(true, properties[0].unique)
        assert_equal("filesize", properties[1].name)
        assert_equal(Rast::PROPERTY_TYPE_UINT, properties[1].type)
        assert_equal(true, properties[1].search)
        assert_equal(false, properties[1].text_search)
        assert_equal(false, properties[1].full_text_search)
        assert_equal(false, properties[1].unique)
        assert_equal("date", properties[2].name)
        assert_equal(Rast::PROPERTY_TYPE_DATE, properties[2].type)
        assert_equal(true, properties[2].search)
        assert_equal(false, properties[2].text_search)
        assert_equal(false, properties[2].full_text_search)
        assert_equal(false, properties[2].unique)
        assert_equal(3, properties.length)
      end
      assert_equal(true, btree_db_exist?(db_name, "text.ngm"))
      assert_equal(true, btree_db_exist?(db_name, "text.rng"))
      assert_equal(true, file_exist?(db_name, "text.pos"))
      assert_equal(true, btree_db_exist?(db_name, "properties.db"))
      assert_equal(true, property_btree_db_exist?(db_name, "filename.inv"))
      assert_equal(true, property_btree_db_exist?(db_name, "filename.ngm"))
      assert_equal(true, property_btree_db_exist?(db_name, "filename.rng"))
      assert_equal(true, property_file_exist?(db_name, "filename.pos"))
      assert_equal(true, property_btree_db_exist?(db_name, "filesize.inv"))
      assert_equal(true, property_btree_db_exist?(db_name, "date.inv"))
      assert_equal(false, dir_exist?(db_name, "text"))
      check_doc_info(db_name, 0, 0)

      options = {
        "encoding" => "utf8",
        "preserve_text" => true,
        "properties" => [
          {
            "name" => "filesize",
            "type" => Rast::PROPERTY_TYPE_UINT,
            "search" => true,
            "text_search" => true,
            "full_text_search" => false,
            "unique" => false,
          },
        ],
      }
      db_name = generate_db_name
      assert_raise(RastError) do
        DB.create(db_name, options)
      end

      options = {
        "encoding" => "utf8",
        "preserve_text" => true,
        "properties" => [
          {
            "name" => "filesize",
            "type" => Rast::PROPERTY_TYPE_UINT,
            "search" => true,
            "text_search" => false,
            "full_text_search" => true,
            "unique" => false,
          },
        ],
      }
      db_name = generate_db_name
      assert_raise(RastError) do
        DB.create(db_name, options)
      end

      options = {
        "encoding" => "utf8",
        "preserve_text" => true,
        "properties" => [
          {
            "name" => "filesize",
            "type" => Rast::PROPERTY_TYPE_UINT,
            "search" => false,
            "text_search" => false,
            "full_text_search" => false,
            "unique" => true,
          },
        ],
      }
      db_name = generate_db_name
      assert_raise(RastError) do
        DB.create(db_name, options)
      end

      options = {
        "byte_order" => Rast::LITTLE_ENDIAN,
        "encoding" => "utf8",
        "preserve_text" => true,
        "properties" => [
          {
            "name" => "title",
            "type" => Rast::PROPERTY_TYPE_STRING,
            "search" => true,
            "text_search" => true,
            "full_text_search" => true,
            "unique" => false,
          },
        ],
      }
      db_path = generate_db_name
      Dir.chdir(File.dirname(db_path))
      db_name = File.basename(db_path)
      DB.create(db_name, options)
      DB.open(db_name, Rast::DB::RDONLY) do |db|
        properties = db.properties
        assert_equal(Rast::LITTLE_ENDIAN, db.byte_order)
        assert_equal("utf8", db.encoding)
        assert_equal("title", properties[0].name)
        assert_equal(Rast::PROPERTY_TYPE_STRING, properties[0].type)
        assert_equal(true, properties[0].search)
        assert_equal(true, properties[0].text_search)
        assert_equal(true, properties[0].full_text_search)
        assert_equal(false, properties[0].unique)
        assert_equal(1, properties.length)
      end
      assert_equal(true, btree_db_exist?(db_name, "text.ngm"))
      assert_equal(true, btree_db_exist?(db_name, "text.rng"))
      assert_equal(true, file_exist?(db_name, "text.pos"))
      assert_equal(true, file_exist?(db_name, "text.pfl"))
      assert_equal(true, btree_db_exist?(db_name, "properties.db"))
      assert_equal(true, property_btree_db_exist?(db_name, "title.inv"))
      assert_equal(true, property_btree_db_exist?(db_name, "title.ngm"))
      assert_equal(true, property_btree_db_exist?(db_name, "title.rng"))
      assert_equal(true, property_file_exist?(db_name, "title.pos"))
      assert_equal(true, property_file_exist?(db_name, "title.pfl"))
      assert_equal(true, dir_exist?(db_name, "text"))
      check_doc_info(db_name, 0, 0)
    end

    def test_open
      options = {
        "encoding" => "utf8",
        "preserve_text" => true,
        "properties" => [],
      }
      db_name = generate_db_name
      DB.create(db_name, options)
      DB.open(db_name, Rast::DB::RDWR) do |db|
        assert_nothing_raised do
          db.register("", {})
        end
      end
      DB.open(db_name, Rast::DB::RDONLY) do |db|
        assert_raise(RastError) do
          db.register("", {})
        end
      end
      options = {
        "sync_threshold_chars" => 10000
      }
      DB.open(db_name, Rast::DB::RDWR, options) do |db|
        assert_equal(10000, db.sync_threshold_chars)
      end
    end

    def test_register
      register_test_empty
      register_test_no_text
      register_test_text_without_normalize
      register_test_text_with_no_change_byte_normalize
      register_test_text_with_change_byte_normalize
      register_test_text_with_null_character
      register_test_property_enable_all_search
      register_test_property_enable_search
      register_test_property_enable_text_search
      register_test_property_enable_full_text_search
      register_test_property_unique
      register_test_summary_text
    end

    def register_test_empty
      options = {
        "encoding" => "utf8",
        "preserve_text" => true,
        "properties" => [],
      }
      db_name = generate_db_name
      DB.create(db_name, options)
      DB.open(db_name) do |db|
        assert_equal(1, db.register("", {}))
      end
      BDB::Btree.open(File.join(db_name, "properties.db"), nil, 0) do |db|
        s = db[[1].pack("I")]
        delete_flag = s.slice!(0, 1).unpack("C")[0]
        num_chars = s.slice!(0, 4).unpack("I")[0]
        
        assert_equal(0, delete_flag)
        assert_equal(0, num_chars)
        assert_equal(0, s.length)
      end
      check_doc_info(db_name, 1, 1)
    end
      
    def register_test_no_text
      options = {
        "encoding" => "utf8",
        "preserve_text" => true,
        "properties" => [
          {
            "name" => "title",
            "type" => Rast::PROPERTY_TYPE_STRING,
            "search" => false,
            "text_search" => false,
            "full_text_search" => false,
          },
          {
            "name" => "modified",
            "type" => Rast::PROPERTY_TYPE_DATE,
            "search" => false,
            "text_search" => false,
            "full_text_search" => false,
          },
          {
            "name" => "serial",
            "type" => Rast::PROPERTY_TYPE_UINT,
            "search" => false,
            "text_search" => false,
            "full_text_search" => false,
          },
        ],
      }
      db_name = generate_db_name
      DB.create(db_name, options)
      DB.open(db_name) do |db|
        properties = {
          "title" => "test",
          "modified" => "2005-02-17T19:04:38",
          "serial" => 10,
        }
        assert_equal(1, db.register("", properties))
      end
      BDB::Btree.open(File.join(db_name, "properties.db"), nil, 0) do |db|
        s = db[[1].pack("I")]
        delete_flag = s.slice!(0, 1).unpack("C")[0]
        num_chars = s.slice!(0, 4).unpack("I")[0]
        title_nbytes = s.slice!(0, 4).unpack("I")[0]
        title = s.slice!(0, title_nbytes)
        modified_nbytes = s.slice!(0, 4).unpack("I")[0]
        modified = s.slice!(0, modified_nbytes)
        serial = s.slice!(0, 4).unpack("I")[0]
        
        assert_equal(0, delete_flag)
        assert_equal(0, num_chars)
        assert_equal("test", title)
        assert_equal("2005-02-17T19:04:38", modified)
        assert_equal(10, serial)
        assert_equal(0, s.length)
      end
      check_doc_info(db_name, 1, 1)
    end

    def register_test_text_without_normalize
      options = {
        "encoding" => "utf8",
        "preserve_text" => true,
        "properties" => [],
      }
      db_name = generate_db_name
      DB.create(db_name, options)
      DB.open(db_name) do |db|
        assert_equal(1, db.register("あいうえお", {}))
      end
      DB.open(db_name) do |db|
        assert_equal("あいうえお", db.get_text(1))
      end
      BDB::Btree.open(File.join(db_name, "properties.db"), nil, 0) do |db|
        s = db[[1].pack("I")]
        delete_flag = s.slice!(0, 1).unpack("C")[0]
        num_chars = s.slice!(0, 4).unpack("I")[0]
        
        assert_equal(0, delete_flag)
        assert_equal(5, num_chars)
        assert_equal(0, s.length)
      end
      check_doc_info(db_name, 1, 1)
    end

    def register_test_text_with_change_byte_normalize
      options = {
        "encoding" => "utf8",
        "preserve_text" => true,
        "properties" => [],
      }
      db_name = generate_db_name
      DB.create(db_name, options)
      DB.open(db_name) do |db|
        assert_equal(1, db.register("あい  Ue\n\r\fO", {}))
      end
      DB.open(db_name) do |db|
        assert_equal("あい Ue O", db.get_text(1))
      end
      check_doc_property(db_name, 1, 0, 7)
      check_doc_info(db_name, 1, 1)
    end
    
    def register_test_text_with_no_change_byte_normalize
      options = {
        "encoding" => "utf8",
        "preserve_text" => true,
        "properties" => [],
      }
      db_name = generate_db_name
      DB.create(db_name, options)
      DB.open(db_name) do |db|
        assert_equal(1, db.register("AIUEO", {}))
        assert_equal(2, db.register("QwErtY", {}))
        assert_equal(3, db.register("あいうZxCv", {}))
      end
      check_doc_property(db_name, 1, 0, 5)
      check_doc_property(db_name, 2, 0, 6)
      check_doc_property(db_name, 3, 0, 7)
      check_registered_text_index(File.join(db_name, "text"), "aiueo")
      check_registered_text_index(File.join(db_name, "text"), "qwerty")
      check_registered_text_index(File.join(db_name, "text"), "あいうzxcv")
      DB.open(db_name) do |db|
        assert_equal("AIUEO", db.get_text(1))
        assert_equal("QwErtY", db.get_text(2))
        assert_equal("あいうZxCv", db.get_text(3))
      end
    end

    def register_test_text_with_null_character
      options = {
        "encoding" => "utf8",
        "preserve_text" => true,
        "properties" => [],
      }
      db_name = generate_db_name
      DB.create(db_name, options)
      DB.open(db_name) do |db|
        assert_equal(1, db.register("abc\0def", {}))
      end
      DB.open(db_name) do |db|
        assert_equal("abc\0def", db.get_text(1))
      end
      check_doc_property(db_name, 1, 0, 7)
      check_doc_info(db_name, 1, 1)
    end

    def register_test_property_enable_all_search
      options = {
        "encoding" => "utf8",
        "preserve_text" => true,
        "properties" => [
          {
            "name" => "string",
            "type" => Rast::PROPERTY_TYPE_STRING,
            "search" => true,
            "text_search" => true,
            "full_text_search" => true,
          },
          {
            "name" => "date",
            "type" => Rast::PROPERTY_TYPE_DATE,
            "search" => true,
            "text_search" => false,
            "full_text_search" => false,
          },
          {
            "name" => "number",
            "type" => Rast::PROPERTY_TYPE_UINT,
            "search" => true,
            "text_search" => false,
            "full_text_search" => false,
          },
         ],
      }
      db_name = generate_db_name
      DB.create(db_name, options)
      DB.open(db_name) do |db|
        text = "aiueo"
        properties = {
          "string" => "abc  DE    F g",
          "date" => "20050101T12:00:00",
          "number" => 100000,
        }
        assert_equal(1, db.register(text, properties))
      end
      check_doc_property(db_name, 1, 0, 5)
      check_summary_text(db_name, 1, "aiueo")

      # search
      check_property_inv(db_name, "string", "abc  DE    F g", 1)
      check_property_inv(db_name, "date", "20050101T12:00:00", 1)
      check_property_inv(db_name, "number", [100000].pack("I"), 1)

      # text_search
      check_registered_text_index(File.join(db_name, "properties", "string"), 
                                  "abc de f g")
      check_not_exist_text_index(db_name, File.join("properties", "date"))
      check_not_exist_text_index(db_name, File.join("properties", "number"))
      
      # full_text_search
      check_registered_text_index(File.join(db_name, "text"), "aiueo")
      check_registered_text_index(File.join(db_name, "text"), "abc de f g")
      check_not_registered_text_index(File.join(db_name, "text"), "eoa")
      check_not_registered_text_index(File.join(db_name, "text"),
                                      "20050101t12:00:00")
      check_not_registered_text_index(File.join(db_name, "text"), "100000")
    end

    def register_test_property_enable_search
      options = {
        "encoding" => "utf8",
        "preserve_text" => true,
        "properties" => [
          {
            "name" => "string",
            "type" => Rast::PROPERTY_TYPE_STRING,
            "search" => true,
            "text_search" => false,
            "full_text_search" => false,
          },
          {
            "name" => "date",
            "type" => Rast::PROPERTY_TYPE_DATE,
            "search" => true,
            "text_search" => false,
            "full_text_search" => false,
          },
          {
            "name" => "number",
            "type" => Rast::PROPERTY_TYPE_UINT,
            "search" => true,
            "text_search" => false,
            "full_text_search" => false,
          },
         ],
      }
      db_name = generate_db_name
      DB.create(db_name, options)
      DB.open(db_name) do |db|
        text = "aiueo"
        properties = {
          "string" => "abc  DE    F g",
          "date" => "20050101T12:00:00",
          "number" => 100000,
        }
        assert_equal(1, db.register(text, properties))
      end
      check_doc_property(db_name, 1, 0, 5)
      check_summary_text(db_name, 1, "aiueo")

      # search
      check_property_inv(db_name, "string", "abc  DE    F g", 1)
      check_property_inv(db_name, "date", "20050101T12:00:00", 1)
      check_property_inv(db_name, "number", [100000].pack("I"), 1)

      # text_search
      check_not_exist_text_index(db_name, File.join("properties", "string"))
      check_not_exist_text_index(db_name, File.join("properties", "date"))
      check_not_exist_text_index(db_name, File.join("properties", "number"))
      
      # full_text_search
      check_registered_text_index(File.join(db_name, "text"), "aiueo")
      check_not_registered_text_index(File.join(db_name, "text"), "abc de f g")
      check_not_registered_text_index(File.join(db_name, "text"),
                                      "20050101t12:00:00")
      check_not_registered_text_index(File.join(db_name, "text"), "100000")
    end

    def register_test_property_enable_text_search
      options = {
        "encoding" => "utf8",
        "preserve_text" => true,
        "properties" => [
          {
            "name" => "string",
            "type" => Rast::PROPERTY_TYPE_STRING,
            "search" => false,
            "text_search" => true,
            "full_text_search" => false,
          },
          {
            "name" => "date",
            "type" => Rast::PROPERTY_TYPE_DATE,
            "search" => false,
            "text_search" => false,
            "full_text_search" => false,
          },
          {
            "name" => "number",
            "type" => Rast::PROPERTY_TYPE_UINT,
            "search" => false,
            "text_search" => false,
            "full_text_search" => false,
          },
         ],
      }
      db_name = generate_db_name
      DB.create(db_name, options)
      DB.open(db_name) do |db|
        text = "aiueo"
        properties = {
          "string" => "aBc  def",
          "date" => "2005-02-21T00:00:00",
          "number" => 100,
        }
        assert_equal(1, db.register(text, properties))
      end
      check_doc_property(db_name, 1, 0, 5)
      check_summary_text(db_name, 1, "aiueo")

      # search
      assert_equal(false, property_btree_db_exist?(db_name, "string"))
      assert_equal(false, property_btree_db_exist?(db_name, "date"))
      assert_equal(false, property_btree_db_exist?(db_name, "number"))

      # text_search
      check_registered_text_index(File.join(db_name, "properties", "string"), 
                                  "abc def")
      
      # full_text_search
      check_registered_text_index(File.join(db_name, "text"), "aiueo")
      check_not_registered_text_index(File.join(db_name, "text"), "abc def")
      check_not_registered_text_index(File.join(db_name, "text"),
                                      "a b c")
      check_not_registered_text_index(File.join(db_name, "text"), "100")
    end

    def register_test_property_enable_full_text_search
      options = {
        "encoding" => "utf8",
        "preserve_text" => true,
        "properties" => [
          {
            "name" => "string",
            "type" => Rast::PROPERTY_TYPE_STRING,
            "search" => false,
            "text_search" => false,
            "full_text_search" => true,
          },
          {
            "name" => "date",
            "type" => Rast::PROPERTY_TYPE_DATE,
            "search" => false,
            "text_search" => false,
            "full_text_search" => false,
          },
          {
            "name" => "number",
            "type" => Rast::PROPERTY_TYPE_UINT,
            "search" => false,
            "text_search" => false,
            "full_text_search" => false,
          },
         ],
      }
      db_name = generate_db_name
      DB.create(db_name, options)
      DB.open(db_name) do |db|
        text = "aiueo"
        properties = {
          "string" => "aBc  def",
          "date" => "20050101T12:00:00",
          "number" => 100000,
        }
        assert_equal(1, db.register(text, properties))
      end
      check_doc_property(db_name, 1, 0, 5)
      check_summary_text(db_name, 1, "aiueo")

      # search
      assert_equal(false, property_btree_db_exist?(db_name, "string"))
      assert_equal(false, property_btree_db_exist?(db_name, "date"))
      assert_equal(false, property_btree_db_exist?(db_name, "number"))

      # text_search
      check_not_exist_text_index(db_name, File.join("properties", "string"))
      check_not_exist_text_index(db_name, File.join("properties", "date"))
      check_not_exist_text_index(db_name, File.join("properties", "number"))
      
      # full_text_search
      check_registered_text_index(File.join(db_name, "text"), "aiueo")
      check_registered_text_index(File.join(db_name, "text"), "abc")
      check_registered_text_index(File.join(db_name, "text"), "abc def")
      check_not_registered_text_index(File.join(db_name, "text"), "eoa")

      options = {
        "encoding" => "utf8",
        "preserve_text" => true,
        "properties" => [
          {
            "name" => "string",
            "type" => Rast::PROPERTY_TYPE_STRING,
            "search" => false,
            "text_search" => false,
            "full_text_search" => true,
          },
          {
            "name" => "date",
            "type" => Rast::PROPERTY_TYPE_DATE,
            "search" => false,
            "text_search" => false,
            "full_text_search" => false,
          },
          {
            "name" => "number",
            "type" => Rast::PROPERTY_TYPE_UINT,
            "search" => false,
            "text_search" => false,
            "full_text_search" => false,
          },
         ],
      }
      db_name = generate_db_name
      DB.create(db_name, options)
      DB.open(db_name) do |db|
        text = ".N ."
        properties = {
          "string" => "NET",
          "date" => "20050101T12:00:00",
          "number" => 100000,
        }
        assert_equal(1, db.register(text, properties))
      end
      check_doc_property(db_name, 1, 0, 4)
      check_summary_text(db_name, 1, ".N .")

      # search
      assert_equal(false, property_btree_db_exist?(db_name, "string"))
      assert_equal(false, property_btree_db_exist?(db_name, "date"))
      assert_equal(false, property_btree_db_exist?(db_name, "number"))

      # text_search
      check_not_exist_text_index(db_name, File.join("properties", "string"))
      check_not_exist_text_index(db_name, File.join("properties", "date"))
      check_not_exist_text_index(db_name, File.join("properties", "number"))
      
      # full_text_search
      check_registered_text_index(File.join(db_name, "text"), ".n .")
      check_registered_text_index(File.join(db_name, "text"), "net")
      check_not_registered_text_index(File.join(db_name, "text"), ".net")
    end
    
    def register_test_property_unique
      options = {
        "encoding" => "utf8",
        "preserve_text" => true,
        "properties" => [
          {
            "name" => "title",
            "type" => Rast::PROPERTY_TYPE_STRING,
            "search" => true,
            "text_search" => false,
            "full_text_search" => false,
            "unique" => false,
          },
          {
            "name" => "filename",
            "type" => Rast::PROPERTY_TYPE_STRING,
            "search" => true,
            "text_search" => false,
            "full_text_search" => false,
            "unique" => true,
          },
         ],
      }
      db_name = generate_db_name
      DB.create(db_name, options)

      DB.open(db_name) do |db|
        text = "hello"
        properties = {
          "title" => "hello",
          "filename" => "hello.txt",
        }
        assert_equal(1, db.register(text, properties))
      end
      check_doc_info(db_name, 1, 1)
      check_doc_property(db_name, 1, 0, 5)
      check_summary_text(db_name, 1, "hello")

      # search
      check_property_inv(db_name, "title", "hello", 1)
      check_property_inv(db_name, "filename", "hello.txt", 1)

      DB.open(db_name) do |db|
        text ="hello"
        properties = {
          "title" => "hello world",
          "filename" => "hello.txt",
        }
        assert_raise(RastError) do
          db.register(text, properties)
        end
      end
      check_doc_info(db_name, 1, 1)

      # search
      check_property_inv(db_name, "title", "hello", 1)
      check_property_inv(db_name, "filename", "hello.txt", 1)
      check_property_inv(db_name, "title", "hello world")

      DB.open(db_name) do |db|
        text = "hello"
        properties = {
          "title" => "hello",
          "filename" => "hello2.txt",
        }
        db.register(text, properties)
      end
      check_doc_info(db_name, 2, 2)
      check_doc_property(db_name, 2, 0, 5)
      check_summary_text(db_name, 2, "hello")

      # search
      check_property_inv(db_name, "title", "hello", 1, 2)
      check_property_inv(db_name, "filename", "hello2.txt", 2)

      options = {
        "encoding" => "utf8",
        "preserve_text" => true,
        "properties" => [
          {
            "name" => "title",
            "type" => Rast::PROPERTY_TYPE_STRING,
            "search" => true,
            "text_search" => false,
            "full_text_search" => false,
            "unique" => true,
          },
          {
            "name" => "filename",
            "type" => Rast::PROPERTY_TYPE_STRING,
            "search" => true,
            "text_search" => false,
            "full_text_search" => false,
            "unique" => true,
          },
         ],
      }
      db_name = generate_db_name
      DB.create(db_name, options)
      DB.open(db_name) do |db|
        text ="hello"
        properties = {
          "title" => "hello",
          "filename" => "hello.txt",
        }
        assert_equal(1, db.register(text, properties))
        properties = {
          "title" => "ruby",
          "filename" => "hello.txt",
        }
        assert_raise(RastError) do
          db.register(text, properties)
        end
      end
      check_doc_info(db_name, 1, 1)
      
      # search
      check_property_inv(db_name, "title", "hello", 1)
      check_property_inv(db_name, "filename", "hello.txt", 1)
      check_property_inv(db_name, "title", "ruby")
    end

    def register_test_summary_text
      options = {
        "encoding" => "utf8",
        "preserve_text" => true,
        "properties" => [],
      }
      db_name = generate_db_name
      DB.create(db_name, options)
      File.open(File.expand_path("doc_info", db_name), "w") do |f|
        f.write([0xAA, 0xAA].pack("I*"))
      end
      DB.open(db_name) do |db|
        assert_equal(0xAA + 1, db.register("aiueo", {}))
      end
      check_doc_info(db_name, 0xAB, 0xAB)
      check_summary_text(db_name, 0xAB, "aiueo")
      File.open(File.expand_path("doc_info", db_name), "w") do |f|
        f.write([127, 127].pack("I*"))
      end
      DB.open(db_name) do |db|
        assert_equal(127 + 1, db.register("aiueo", {}))
      end
      check_doc_info(db_name, 128, 128)
      check_summary_text(db_name, 128, "aiueo")
      DB.open(db_name) do |db|
        assert_equal(127 + 2, db.register("aiueo", {}))
      end
      check_doc_info(db_name, 129, 129)
      check_summary_text(db_name, 129, "aiueo")
    end

    def test_sync
      options = {
        "encoding" => "utf8",
        "preserve_text" => true,
        "properties" => [],
      }
      db_name = File.expand_path("testdb", @db_dir)
      DB.create(db_name, options)
      DB.open(db_name) do |db|
        db.register("本日は晴天なり", {})
        db.sync
        result = db.search("晴天")
        assert_equal(1, result.hit_count)
      end
    end

    def test_search
      search_test_simple
      search_test_summary
      search_test_text_and_property
      search_test_property_and_summary
      search_test_sort_with_score
      search_test_calc_score
      search_test_sort_with_property
      search_test_set_range
    end

    def search_test_simple
      options = {
        "encoding" => "utf8",
        "preserve_text" => true,
        "properties" => [],
      }
      db_name = generate_db_name
      DB.create(db_name, options)
      DB.open(db_name) do |db|
        db.register("昨日は雨天なり", {})
        db.register("本日は晴天なり", {})
        db.register("明日は雪っす", {})
      end
      DB.open(db_name) do |db|
        result = db.search("晴天")
        assert_equal(1, result.hit_count)
        assert_equal("晴天", result.terms[0].term)
        assert_equal(1, result.terms[0].doc_count)
        assert_equal(1, result.terms.length)
        assert_equal(2, result.items[0].doc_id)
        assert_equal(0, result.items[0].properties.length)
        assert_equal(1, result.items.length)
      end
    end

    def search_test_summary
      options = {
        "encoding" => "utf8",
        "preserve_text" => true,
        "properties" => [
          {
            "name" => "title",
            "type" => Rast::PROPERTY_TYPE_STRING,
            "search" => false,
            "text_search" => false,
            "full_text_search" => true,
            "unique" => false,
          },
        ],
      }
      db_name = generate_db_name
      DB.create(db_name, options)
      DB.open(db_name) do |db|
        db.register("0123456789", {"title" => "full text search property"})
        db.register("零一二三四五六七八九",
                    {"title" => "full text search property"})
        db.register("零1二3四5六7八9",
                    {"title" => "full text search property"})
      end
      DB.open(db_name) do |db|
        search_options = {"need_summary" => true, "summary_nchars" => 6}
        result = db.search("5", search_options)
        assert_equal(1, result.items[0].doc_id)
        assert_equal("345678", result.items[0].summary)
        assert_equal(3, result.items[1].doc_id)
        assert_equal("3四5六7八", result.items[1].summary)
        assert_equal(2, result.items.length)

        result = db.search("1", search_options)
        assert_equal(1, result.items[0].doc_id)
        assert_equal("012345", result.items[0].summary)
        assert_equal(3, result.items[1].doc_id)
        assert_equal("零1二3四5", result.items[1].summary)
        assert_equal(2, result.items.length)

        result = db.search("8", search_options)
        assert_equal(1, result.items[0].doc_id)
        assert_equal("456789", result.items[0].summary)
        assert_equal(1, result.items.length)

        result = db.search("五", search_options)
        assert_equal(2, result.items[0].doc_id)
        assert_equal("三四五六七八", result.items[0].summary)
        assert_equal(1, result.items.length)

        result = db.search("一", search_options)
        assert_equal(2, result.items[0].doc_id)
        assert_equal("零一二三四五", result.items[0].summary)
        assert_equal(1, result.items.length)

        result = db.search("八", search_options)
        assert_equal(2, result.items[0].doc_id)
        assert_equal("四五六七八九", result.items[0].summary)
        assert_equal(3, result.items[1].doc_id)
        assert_equal("四5六7八9", result.items[1].summary)
        assert_equal(2, result.items.length)

        result = db.search("full", search_options)
        assert_equal(1, result.items[0].doc_id)
        assert_equal("012345", result.items[0].summary)
        assert_equal(2, result.items[1].doc_id)
        assert_equal("零一二三四五", result.items[1].summary)
        assert_equal(3, result.items[2].doc_id)
        assert_equal("零1二3四5", result.items[2].summary)
        assert_equal(3, result.items.length)

        result = db.search("12", search_options)
        assert_equal(1, result.items[0].doc_id)
        assert_equal("012345", result.items[0].summary)
        assert_equal(1, result.items.length)

        result = db.search("五六", search_options)
        assert_equal(2, result.items[0].doc_id)
        assert_equal("三四五六七八", result.items[0].summary)
        assert_equal(1, result.items.length)

        result = db.search("0123456789", search_options)
        assert_equal(1, result.items[0].doc_id)
        assert_equal("234567", result.items[0].summary)
        assert_equal(1, result.items.length)

        search_options = {"need_summary" => true, "summary_nchars" => 11}
        result = db.search("5", search_options)
        assert_equal(1, result.items[0].doc_id)
        assert_equal("0123456789", result.items[0].summary)
        assert_equal(3, result.items[1].doc_id)
        assert_equal("零1二3四5六7八9", result.items[1].summary)
        assert_equal(2, result.items.length)

        search_options = {"need_summary" => true, "summary_nchars" => 12}
        result = db.search("5", search_options)
        assert_equal(1, result.items[0].doc_id)
        assert_equal("0123456789", result.items[0].summary)
        assert_equal(3, result.items[1].doc_id)
        assert_equal("零1二3四5六7八9", result.items[1].summary)
        assert_equal(2, result.items.length)
      end
    end

    def search_test_text_and_property
      options = {
        "encoding" => "utf8",
        "preserve_text" => false,
        "properties" => [
          {
            "name" => "filename",
            "type" => Rast::PROPERTY_TYPE_STRING,
            "search" => false,
            "text_search" => false,
            "full_text_search" => true,
          },
        ],
      }

      db_name = generate_db_name
      DB.create(db_name, options)
      DB.open(db_name) do |db|
        db.register("bcxy abc", {"filename" => "xyz"})
        db.register("bcxy", {"filename" => "123"})
      end
      DB.open(db_name) do |db|
        result = db.search("abc")
        assert_equal(1, result.hit_count)
        result = db.search("xyz")
        assert_equal(1, result.hit_count)
        result = db.search("abcxyz")
        assert_equal(0, result.hit_count)
        result = db.search('"abc xyz"')
        assert_equal(0, result.hit_count)
      end
    end

    def search_test_property_and_summary
      options = {
        "encoding" => "utf8",
        "preserve_text" => true,
        "properties" => [
          {
            "name" => "filename",
            "type" => Rast::PROPERTY_TYPE_STRING,
            "search" => false,
            "text_search" => false,
            "full_text_search" => false,
          },
        ],
      }

      db_name = generate_db_name
      DB.create(db_name, options)
      DB.open(db_name) do |db|
        db.register("昨日は雨天なり", {"filename" => "ame.txt"})
        db.register("a" * 100 + "晴天" + "b" * 100,
                    {"filename" => "hare.txt"})
        db.register("明日は雪っす", {"filename" => "yuki.txt"})
        db.register("", {"filename" => "yuki.txt"})
      end
      DB.open(db_name) do |db|
        result = db.search("晴天",
                           {
                             "properties" => ["filename"],
                             "need_summary" => true,
                           })
        assert_equal(1, result.hit_count)
        assert_equal("晴天", result.terms[0].term)
        assert_equal(1, result.terms[0].doc_count)
        assert_equal(1, result.terms.length)
        assert_equal(2, result.items[0].doc_id)
        assert_equal("hare.txt", result.items[0].properties[0])
        assert_equal(1, result.items[0].properties.length)
        expected_summary = "a" * 49 + "晴天" + "b" * 49
        assert_equal(expected_summary, result.items[0].summary)
        assert_equal(1, result.items.length)
      end

      db_name = generate_db_name
      DB.create(db_name, options)
      DB.open(db_name) do |db|
        db.register("1" + "abc" * 100, {"filename" => "abc.txt"})
      end
      DB.open(db_name) do |db|
        result = db.search("abc",
                           {
                             "summary_nchars" => 301,
                             "need_summary" => true,
                           })
        assert_equal(1, result.hit_count)
        assert_equal("abc", result.terms[0].term)
        assert_equal(1, result.terms[0].doc_count)
        assert_equal(1, result.terms.length)
        assert_equal(1, result.items[0].doc_id)
        assert_equal("1" + "abc" * 100, result.items[0].summary)
        assert_equal(1, result.items.length)
      end
    end

    def search_test_sort_with_score
      options = {
        "encoding" => "utf8",
        "properties" => [],
      }
      db_name = generate_db_name
      DB.create(db_name, options)
      DB.open(db_name) do |db|
        db.register("あい" * 99 + "晴天", {}) 
        db.register("晴天" * 100, {})
        db.register("あい" * 49 + "晴天", {}) 
     end
      DB.open(db_name) do |db|
        # descend
        result = db.search("晴天")
        assert_equal(3, result.hit_count)
        assert_equal("晴天", result.terms[0].term)
        assert_equal(3, result.terms[0].doc_count)
        assert_equal(1, result.terms.length)
        assert_equal(2, result.items[0].doc_id)
        assert_equal(3, result.items[1].doc_id)
        assert_equal(1, result.items[2].doc_id)
        assert_equal(3, result.items.length)

        # ascend
        search_options = {
          "sort_order" => Rast::SORT_ORDER_ASCENDING
        }
        result = db.search("晴天", search_options)
        assert_equal(3, result.hit_count)
        assert_equal("晴天", result.terms[0].term)
        assert_equal(3, result.terms[0].doc_count)
        assert_equal(1, result.terms.length)
        assert_equal(1, result.items[0].doc_id)
        assert_equal(3, result.items[1].doc_id)
        assert_equal(2, result.items[2].doc_id)
        assert_equal(3, result.items.length)
      end
    end

    def search_test_calc_score
      options = {
        "encoding" => "utf8",
        "properties" => [
          {
            "name" => "filename",
            "type" => Rast::PROPERTY_TYPE_STRING,
            "search" => true,
            "text_search" => false,
            "full_text_search" => false,
          },
        ],
      }
      db_name = generate_db_name
      DB.create(db_name, options)
      DB.open(db_name) do |db|
        db.register("aaa" * 100, {"filename" => "abc.txt"})
     end
      DB.open(db_name) do |db|
        options = {
          "score_method" => Rast::SCORE_METHOD_NONE,
        }
        result = db.search("aaa", options)
        assert_equal(0, result.items[0].score)

        options = {
          "score_method" => Rast::SCORE_METHOD_TFIDF,
        }
        result = db.search("aaa", options)
        assert_operator(0, :<, result.items[0].score)

        result = db.search("aaa", {})
        assert_operator(0, :<, result.items[0].score)

        options = {
          "sort_method" => Rast::SORT_METHOD_PROPERTY,
          "sort_property" => "filename"
        }
        result = db.search("aaa", options)
        assert_operator(0, :<, result.items[0].score)

        options = {
          "score_method" => Rast::SCORE_METHOD_TFIDF,
          "sort_method" => Rast::SORT_METHOD_PROPERTY,
          "sort_property" => "filename"
        }
        result = db.search("aaa", options)
        assert_operator(0, :<, result.items[0].score)
      end
    end

    def search_test_sort_with_property
      options = {
        "encoding" => "utf8",
        "properties" => [
          {
            "name" => "filename",
            "type" => Rast::PROPERTY_TYPE_STRING,
            "search" => false,
            "text_search" => false,
            "full_text_search" => false,
          },
          {
            "name" => "serial",
            "type" => Rast::PROPERTY_TYPE_UINT,
            "search" => false,
            "text_search" => false,
            "full_text_search" => false,
          },
        ],
      }
      db_name = generate_db_name
      DB.create(db_name, options)
      DB.open(db_name) do |db|
        db.register("昨日は雨天なり",
                    {"filename" => "2.txt", "serial" => 10})
        db.register("本日は晴天なり",
                    {"filename" => "3.txt", "serial" => 5})
        db.register("明日は雪っすり",
                    {"filename" => "1.txt", "serial" => 15})
      end
      DB.open(db_name) do |db|
        result = db.search("日は",
                           {
                             "sort_method" => Rast::SORT_METHOD_PROPERTY,
                             "sort_property" => "filename",
                             "properties" => ["filename"],
                           })
        assert_equal(3, result.hit_count)
        assert_equal(3, result.terms[0].doc_count)
        assert_equal(1, result.terms.length)
        assert_equal(3, result.items[0].doc_id)
        assert_equal(1, result.items[1].doc_id)
        assert_equal(2, result.items[2].doc_id)
        assert_equal(3, result.items.length)

        result = db.search("日は",
                           {
                             "sort_method" => Rast::SORT_METHOD_PROPERTY,
                             "sort_property" => "filename",
                             "sort_order" => Rast::SORT_ORDER_ASCENDING,
                             "properties" => ["filename"],
                           })
        assert_equal(3, result.hit_count)
        assert_equal(3, result.terms[0].doc_count)
        assert_equal(1, result.terms.length)
        assert_equal(3, result.items[0].doc_id)
        assert_equal(1, result.items[1].doc_id)
        assert_equal(2, result.items[2].doc_id)
        assert_equal(3, result.items.length)

        result = db.search("日は",
                           {
                             "sort_method" => Rast::SORT_METHOD_PROPERTY,
                             "sort_property" => "filename",
                             "sort_order" => Rast::SORT_ORDER_DESCENDING,
                             "properties" => ["filename"],
                           })
        assert_equal(3, result.hit_count)
        assert_equal(3, result.terms[0].doc_count)
        assert_equal(1, result.terms.length)
        assert_equal(2, result.items[0].doc_id)
        assert_equal(1, result.items[1].doc_id)
        assert_equal(3, result.items[2].doc_id)
        assert_equal(3, result.items.length)

        result = db.search("日は",
                           {
                             "sort_method" => Rast::SORT_METHOD_PROPERTY,
                             "sort_property" => "serial",
                             "properties" => ["serial"],
                           })
        assert_equal(3, result.hit_count)
        assert_equal(3, result.terms[0].doc_count)
        assert_equal(1, result.terms.length)
        assert_equal(2, result.items[0].doc_id)
        assert_equal(1, result.items[1].doc_id)
        assert_equal(3, result.items[2].doc_id)
        assert_equal(3, result.items.length)

        result = db.search("日は",
                           {
                             "sort_method" => Rast::SORT_METHOD_PROPERTY,
                             "sort_property" => "serial",
                             "sort_order" => Rast::SORT_ORDER_ASCENDING,
                             "properties" => ["serial"],
                           })
        assert_equal(3, result.hit_count)
        assert_equal(3, result.terms[0].doc_count)
        assert_equal(1, result.terms.length)
        assert_equal(2, result.items[0].doc_id)
        assert_equal(1, result.items[1].doc_id)
        assert_equal(3, result.items[2].doc_id)
        assert_equal(3, result.items.length)

        result = db.search("日は",
                           {
                             "sort_method" => Rast::SORT_METHOD_PROPERTY,
                             "sort_property" => "serial",
                             "sort_order" => Rast::SORT_ORDER_DESCENDING,
                             "properties" => ["serial"],
                           })
        assert_equal(3, result.hit_count)
        assert_equal(3, result.terms[0].doc_count)
        assert_equal(1, result.terms.length)
        assert_equal(3, result.items[0].doc_id)
        assert_equal(1, result.items[1].doc_id)
        assert_equal(2, result.items[2].doc_id)
        assert_equal(3, result.items.length)
      end
    end

    def  search_test_set_range
      options = {
        "encoding" => "utf8",
        "properties" => [],
      }
      db_name = generate_db_name
      DB.create(db_name, options)
      DB.open(db_name) do |db|
        db.register("1", {}) 
        db.register("1", {}) 
        db.register("1", {}) 
        db.register("1", {}) 
        db.register("1", {}) 
     end
      DB.open(db_name) do |db|
        search_options = {
          "start_no" => 0,
          "num_items" => 3,
        }
        result = db.search("1", search_options)
        assert_equal(5, result.hit_count)
        assert_equal(3, result.items.length)
        
        search_options = {
          "start_no" => 1,
          "num_items" => 3,
        }
        result = db.search("1", search_options)
        assert_equal(5, result.hit_count)
        assert_equal(3, result.items.length)

        search_options = {
          "start_no" => 4,
          "num_items" => 10,
        }
        result = db.search("1", search_options)
        assert_equal(5, result.hit_count)
        assert_equal(1, result.items.length)

        search_options = {
          "start_no" => 20,
          "num_items" => 10,
        }
        result = db.search("1", search_options)
        assert_equal(5, result.hit_count)
        assert_equal(0, result.items.length)
      end
    end

    def test_delete
      options = {
        "encoding" => "utf8",
        "preserve_text" => true,
        "properties" => [],
      }
      db_name = generate_db_name
      DB.create(db_name, options)
      DB.open(db_name) do |db|
        db.register("本日は晴天なり", {})
      end
      DB.open(db_name) do |db|
        result = db.search("晴天")
        assert_equal(1, result.hit_count)
        db.delete(1)
      end
      DB.open(db_name) do |db|
        result = db.search("晴天")
        assert_equal(0, result.hit_count)
      end
    end

    def test_update
      options = {
        "encoding" => "utf8",
        "preserve_text" => true,
        "properties" => [],
      }
      db_name = generate_db_name
      DB.create(db_name, options)
      DB.open(db_name) do |db|
        db.register("本日は晴天なり", {})
      end
      DB.open(db_name) do |db|
        result = db.search("晴天")
        assert_equal(1, result.hit_count)
        assert_equal(2, db.update(1, "明日は雪っす", {}))
      end
      DB.open(db_name) do |db|
        result = db.search("晴天")
        assert_equal(0, result.hit_count)
        result = db.search("雪っ")
        assert_equal(1, result.hit_count)
      end
    end

    def test_get_text
      options = {
        "encoding" => "utf8",
        "preserve_text" => false,
        "properties" => [],
      }
      db_name = generate_db_name
      DB.create(db_name, options)
      assert_equal(false, dir_exist?(db_name, "text"))
      DB.open(db_name) do |db|
        db.register("本日は晴天なり", {})
      end
      DB.open(db_name) do |db|
        assert_raise(RastError) do
          db.get_text(1)
        end
      end
    end

    private

    def file_exist?(db_name, filename)
      return File.exist?(File.expand_path(filename, db_name))
    end

    def dir_exist?(db_name, dir_name)
      return File.directory?(File.expand_path(dir_name, db_name)) 
    end

    def btree_db_exist?(db_name, filename)
      begin
        BDB::Btree.open(File.expand_path(filename, db_name)) do |db|
        end
        return true
      rescue
        return false
      end
    end

    def hash_db_exist?(db_name, filename)
      begin
        BDB::Hash.open(File.expand_path(filename, db_name)) do |db|
        end
        return true
      rescue
        return false
      end
    end

    def property_btree_db_exist?(db_name, filename)
      return btree_db_exist?(db_name, File.join("properties", filename))
    end

    def property_file_exist?(db_name, filename)
      return file_exist?(db_name, File.join("properties", filename))
    end

    def check_exist_text_index(db_name, path, is_exist = true)
      assert_equal(is_exist, btree_db_exist?(db_name, path + ".ngm"))
      assert_equal(is_exist, btree_db_exist?(db_name, path + ".rng"))
      assert_equal(is_exist, file_exist?(db_name, path + ".pos"))
      assert_equal(is_exist, file_exist?(db_name, path + ".pfl"))
    end

    def check_not_exist_text_index(db_name, path)
      check_exist_text_index(db_name, path, false)
    end

    def check_doc_info(db_name, expected_max_doc_id, expected_num_doc)
      File.open(File.expand_path("doc_info", db_name)) do |f|
        max_doc_id, num_doc = *f.read.unpack("I*")
        assert_equal(expected_max_doc_id, max_doc_id)
        assert_equal(expected_num_doc, num_doc)
      end
    end

    def check_doc_property(db_name, doc_id,
                           expected_delete_flag, expected_num_chars)
      BDB::Btree.open(File.join(db_name, "properties.db"), nil, 0) do |db|
        s = db[[doc_id].pack("I")]
        delete_flag = s.slice!(0, 1).unpack("C")[0]
        num_chars = s.slice!(0, 4).unpack("I")[0]
        
        assert_equal(expected_delete_flag, delete_flag)
        assert_equal(expected_num_chars, num_chars)
      end
    end

    def check_registered_text_index(text_index_name, text, is_registered = true)
      text_index = TextIndex.new(text_index_name)
      res = text_index.search(text)
      assert_equal(!is_registered, res.candidates.empty?)
    end

    def check_not_registered_text_index(text_index_name, text)
      check_registered_text_index(text_index_name, text, false)
    end
    
    def check_property_inv(db_name, property_name, property_value,
                           *expected_doc_ids)
      path = File.join(db_name, "properties", property_name + ".inv")
      BDB::Btree.open(path, nil, BDB::RDONLY) do |db|
        c = db.cursor
        key, value = c.set(property_value)
        for doc_id in expected_doc_ids
          assert_equal(property_value, key, "property not found")
          assert_equal(doc_id, value.unpack("I*")[0])
          key, value = c.next
        end
        assert_not_equal(property_value, key, "too many hit")
      end
    end
    
    def check_summary_text(db_name, doc_id, text)
      DB.open(db_name) do |db|
        assert_equal(text, db.get_text(doc_id))
      end
      p1, p2 = [doc_id].pack("v").unpack("H2H2")
      path = File.join(db_name, "text", p1.upcase, p2.upcase, doc_id)
      assert_equal(true, File.exist?(path), path)
    end
  end
end
